// ©2014 The Telos Alliance. All Rights Reserved.

// congraph.js: jQuery and flot based graph of connection stats
// author:      Jeff Bencin <jeff.bencin@gmail.com>
//
// REQUIRES:
//   jquery.js
//   jquery.flot.js
//   common.js
//   congraph.css
//
// Javascript namespace: cgr_*
// CSS namespace: cgr-* (CSS in file congraph.css)

// Make sure console.log does not crash in IE
if(!window.console) { window.console = {log: function(){} }; }

// The selector of the element where everything will be done
var cgr_rootSelector = '#jquerySectCongraph';

// Called when document loads
$(document).ready(function() {
	var root = $(cgr_rootSelector);

	// Hide element to prevent reflow before we finish
	root.hide();
	root.empty();

	// Construct graph area
	root.attr('id', 'cgr-root');
	$('<div>').attr('id', 'cgr-graph_messages').appendTo(root);
	$('<div>').attr('id', 'cgr-graph').appendTo(root);
	$('<div>').attr('id', 'cgr-graph_control').appendTo(root);

	// Apply changes
	root.show();

	cgr_initializeGraph();
});

function cgr_initializeGraph() {
	const dataInterval = 1000;  // Interval at which data is refreshed, milliseconds
	const FHGInterval  = 3;     // Interval at which FHG libs report stats, seconds
	const msPlotMin    = 30000; // Minimum amount of time to show on graph
	const pxPerPoint   = 2;     // Minimum pixels per point of data

	var container = $("#cgr-graph");
	var bitrates  = null;  // Array of all possible bitrates

	var datasets = [
		{ label: "Rcv Bitrate",  color: 4, yaxis: 2, data: [], cgr_data: [], cgr_key: "rcv_bitr", cgr_mode: "raw"   },
		{ label: "Rcv Dropped",  color: 1, yaxis: 1, data: [], cgr_data: [], cgr_key: "rcv_drop", cgr_mode: "afhgr" },
		{ label: "Rcv Buffered", color: 2, yaxis: 1, data: [], cgr_data: [], cgr_key: "rcv_buff", cgr_mode: "raw"   },
		{ label: "Rcv Lost",     color: 3, yaxis: 1, data: [], cgr_data: [], cgr_key: "rcv_lost", cgr_mode: "afhgr" },
		{ label: "Rcv Frames",   color: 0, yaxis: 1, data: [], cgr_data: [], cgr_key: "rcv_pkts", cgr_mode: "afhgr" },
		{ label: "Xmt Bitrate",  color: 9, yaxis: 2, data: [], cgr_data: [], cgr_key: "xmt_bitr", cgr_mode: "raw"   },
		{ label: "Xmt Dropped",  color: 6, yaxis: 1, data: [], cgr_data: [], cgr_key: "xmt_drop", cgr_mode: "afhgx" },
		{ label: "Xmt Buffered", color: 7, yaxis: 1, data: [], cgr_data: [], cgr_key: "xmt_buff", cgr_mode: "raw"   },
		{ label: "Xmt Lost",     color: 8, yaxis: 1, data: [], cgr_data: [], cgr_key: "xmt_lost", cgr_mode: "afhgx" },
		{ label: "Xmt Frames",   color: 5, yaxis: 1, data: [], cgr_data: [], cgr_key: "xmt_pkts", cgr_mode: "afhgx" }
	];

	$.each(datasets, function(key, val) {
		// Use different types of plots for different modes
		switch (val.cgr_mode) {
			case "afhgr":
			case "afhgx":
				val.lines = {
					show: true,
					steps: true
				}
				break;
		}
	});

	// Draw graph controls
	var control = $("#cgr-graph_control");
	// Draw checkbox to toggle each dataset on/off
	$.each(datasets, function(key, dataset) {
		var label = dataset.label.split(" ");
		var i = 0;
		var last = label.length - 1;
		var elem = control;
		while (i < last) {
			var fsLabel = label[i++];
			var className = "cgr-fs-" + fsLabel;
			var newElem = elem.find("." + className);
			if (newElem.length===0) {
				newElem = $('<fieldset>').addClass(className);
				$('<legend>').text(fsLabel).appendTo(newElem);
				newElem.appendTo(elem);
			}
			elem = newElem;
		}
		var input = $("<label><input type='checkbox' name='" + key + "'></input>" + label[i] + "</label>");
		input.appendTo(elem);
		if (/bitr/.test(dataset.cgr_key)) {
			input.find('input').prop('checked', true);
		}
	});
	// Draw other options
	var options = $('<fieldset id="cgr-fs-options">').appendTo(control);
	$('<legend>').text("Options").appendTo(options);
	$("<label><input type='checkbox' id='cgr-val-clock_12h'></input> 12h clock </label>").appendTo(options);;

	// Redraw graph when control inputs are changed
	control.find("input").on('change', updateGraphDisplay);

	// Get total amount of time that can be shown on screen
	function getMsPlotMax() {
		var msPlotMax = (container.width() / pxPerPoint) * dataInterval;
		return Math.max(msPlotMin, msPlotMax);
	};

	// Remove elements from array that are too old to be displayed
	function trimOldData (arr, now, span) {
		while (true) {
			var len = arr.length;
			if (!len) break;
			var timeDiff = now - arr[len-1].time;
			if (timeDiff < span) break;
			arr.pop();
		}
	}

	// Process and store data from AJAX request
	function updateGraphData(data) {
		// Get these variables only once per function call
		var timestamp  = Date.now();
		var msPlotMax = getMsPlotMax();

		// Map datasets to variables in JSON structure
		var varMap = {
			rcv_pkts:  {val: data.rcv.pkts.total},
			rcv_drop:  {val: data.rcv.pkts.dropped},
			rcv_buff:  {val: data.rcv.pkts.buffered},
			rcv_lost:  {val: data.rcv.pkts.lost},
			rcv_bitr:  {val: data.rcv.bitrate},
			rcv_mbuf:  {val: data.rcv.minBuff},
			rcv_srat:  {val: data.rcv.sampleRate},
			rcv_chan:  {val: data.rcv.channels},
			rcv_codc:  {val: data.rcv.codecName},
			xmt_pkts:  {val: data.xmt.pkts.total},
			xmt_drop:  {val: data.xmt.pkts.dropped},
			xmt_buff:  {val: data.xmt.pkts.buffered},
			xmt_lost:  {val: data.xmt.pkts.lost},
			xmt_bitr:  {val: data.xmt.bitrate},
			xmt_mbuf:  {val: data.xmt.minBuff},
			xmt_srat:  {val: data.xmt.sampleRate},
			xmt_chan:  {val: data.xmt.channels},
			xmt_codc:  {val: data.xmt.codecName},
			call_peer: {val: data.call.peer},
			call_time: {val: data.call.duration},
			call_stat: {val: data.call.status}
		};

		// Process values and Update text displays of these vars, if they exist
		for (var prop in varMap) {
			var obj = varMap[prop];
			var val = obj.val;
			if (obj.func) {
				val = obj.func(val);
				obj.val = val;
			}
			$('.cgr-disp-' + prop).text(val);
		}

		// Update graph datasets
		$.each(datasets, function(key, dataset) {
			const rawVal = varMap[dataset.cgr_key].val;
			var arr = dataset.cgr_data;
			var val;
			var plot = true; // Set to false to not plot this point

			// Make sure first element is not undefined
			if (!arr.length) {
				arr.unshift({
					time: 0
				});
			}

			switch (dataset.cgr_mode) {
				case "diff": // Show difference from last value
					var lastVal = arr[0].rawVal || rawVal;
					val = rawVal - lastVal;
					break;
				case "adiff": // Average diff over last 6 values for smooth curve
					var diffIdx = Math.min(arr.length-1, 5);
					if (diffIdx < 0) break;

					var diffPoint = arr[diffIdx];
					var diffVal = rawVal - diffPoint.rawVal;
					var diffSec = (timestamp - diffPoint.time) / 1000;
					val = diffVal / diffSec;
					break;
				case "afhgx": // Average over FHGInterval, use xmt timestamp
				case "afhgr": // Average over FHGInterval, use rcv timestamp
					var misc = arr[0].misc || { count: 0, rtime: 0 };
					var remoteTime;
					switch (dataset.cgr_mode) {
						case "afhgx": remoteTime = data.xmt.timestamp; break;
						case "afhgr": remoteTime = data.rcv.timestamp; break;
					}
					if (remoteTime <= misc.rtime) {
						if (++misc.count < 4) {
							plot = false;
							break;
						}
					}
					misc.count = 0;
					var lastVal = arr[0].rawVal || rawVal;
					misc.rtime = remoteTime;
					val = (rawVal - lastVal) / FHGInterval;
					break;
				case "raw": // Show value as received
				default:
					val = rawVal;
					break;
			}

			// Add new data point to array
			if (plot) {
				arr.unshift({
					time:   timestamp,
					val:    val,
					rawVal: rawVal,
					misc:   misc
				});
			}

			trimOldData(arr, timestamp, msPlotMax);
		});
	}

	// Call to redraw graph
	function updateGraphDisplay() {
		// Get these variables only once per function call
		var timestamp = Date.now();
		var msPlotMax = getMsPlotMax();

		// Find selected datasets
		var selectedData = [];
		control.find("input:checked").each(function () {
			var key = $(this).attr("name");
			if (key && datasets[key]) {
				selectedData.push(datasets[key]);
			}
		});

		var maxBr = 0;

		// Generate coordinates only for selected datasets
		$.each(selectedData, function(key, dataset) {
			// Make (x,y) coordinates
			var arr = dataset.cgr_data;

			trimOldData(arr, timestamp, msPlotMax);

			var data = [];
			for (var i in arr) {
				var p = arr[i];
				data.push([p.time, p.val])
			}

			// Find largest bitrate displayed
			if (dataset.yaxis===2) {
				for (var i in arr) {
					maxBr = Math.max(maxBr, arr[i].val);
				}
			}

			dataset.data = data;
		});

		// Calculate ticks on y-axis for bitrates
		var brTicks = [[0, "0 Kbps"]];

		if (bitrates) {
			const maxBrTicks = 7;
			var i;
			for (i in bitrates) {
				if (bitrates[i] > maxBr) {
					break;
				}
			}
			i -= (maxBrTicks-2);

			var j = 1;
			while (j<maxBrTicks) {
				var br = bitrates[i++];
				if (!br) continue;
				brTicks[j++] = [br, (br/1000) + " Kbps"];
			}
		}

		var clock_12h = $("#cgr-val-clock_12h").prop("checked");

		// Configure display
		var options = {
			xaxis: {
				mode: "time",
				minTickSize: [5, "second"],
				tickFormatter: function (val, axis) {
					var d = new Date(val);
					var h = d.getHours();
					if (clock_12h && h > 12) {
						h -= 12;
					}
					var m = d.getMinutes().toString();
					if (m.length===1) {
						m = "0" + m;
					}
					var s = d.getSeconds().toString();
					if (s.length===1) {
						s = "0" + s;
					}
					return [h, m , s].join(":");
				}
			},
			yaxis: {
				min: 0,
				lineHeight: 10
			},
			yaxes: [
				/* Packets */ {
					minTickSize: 10,
					tickDecimals: 0
				},
				/* Bitrates */ {
					ticks: brTicks,
					position: "right"
				}
			],
			legend: {
				position: "nw",
				show: true
			}
		};

		// Draw selected datasets
		container.plot(selectedData, options); 
	}

	// Get scale for y axis
	$.getJSON("json/bitrates", function(data) {
		var brs = com_parseBitratesFromJSON(data);
		// Convert to bits/sec
		for (var i in brs) {
			brs[i] *= 1000;
		}
		bitrates = brs;
	});

	// Set up recurring AJAX request to update graph
	com_ajaxSubscribe("json/connection_stats", dataInterval, function(data) {
		updateGraphData(data);
		updateGraphDisplay();
	});
}
