// ©2014 The Telos Alliance. All Rights Reserved.

// meters.js: Javascript audio meters
// author:    Jeff Bencin <jeff.bencin@gmail.com>
//
// Javascript namespace: meters_*

// Make sure console.log does not crash in IE
if(!window.console) { window.console = {log: function(){} }; }

var g_nsegments      = 86;      // Number of segments per meter
var g_nsrc           = 0;       // Number of source meters
var g_ndst           = 0;       // Number of destination meters
var g_surround       = [];      // Are we in surround mode
var g_VER            = "";      // Store result of VER command

var g_mtr_element;              // Reference to element in which meters are drawn
var g_mtr_request;              // AJAX request object for meters
var g_mtr_conn_err;             // true if cannot connect to server
var g_mtr_interval   = 30;      // How often we poll for response, in ms
var g_mtr_resp_idx   = 0;       // Our current place in g_mtr_request.responseText
var g_mtr_timeout;              // Javascript timeout object
var g_mtr_timeout_ms = 2000;    // Amount of time before connection has timed out, 2 sec
var g_mtr_reconnect  = 200;     // Delay in ms used to reconnect after timeout

var g_lwr_request;              // AJAX request object for lwrp requests
var g_lwr_conn_err;             // true if cannot connect to server
var g_lwr_resp_idx   = 0;       // Our current place in g_lwr_request.responseText
var g_lwr_timeout;              // Javascript timeout object
var g_lwr_timeout_ms = 60000;   // Amount of time before connection has timed out, 60 sec
var g_lwr_reconnect  = 1000;    // Delay in ms used to reconnect after timeout

var g_resp_idx_limit = 500000;  // If responseText exceeds this size, restart connection to clear

// Arrays of elements for each meter segment
// cached, so we don't have to use getElementById() each time
var g_src_M_elems; // mono
var g_src_L_elems;
var g_src_R_elems;
var g_dst_M_elems; // mono
var g_dst_L_elems;
var g_dst_R_elems;

// Also cache the number of segments from last update
// This way we only have to redraw the difference
var g_src_M_seg; // mono
var g_src_L_seg;
var g_src_R_seg;
var g_src_M_seg; // mono
var g_dst_L_seg;
var g_dst_R_seg;

// Number of channels for each src and dst (mono, stereo, or zero if disabled)
var g_nchn = [];

// HTML fragments used in meter tables
var g_mtr_fragments = [];

// Meter segment colors
var g_seg_black  = '#000000';
var g_seg_green  = '#7CFC00';
var g_seg_yellow = '#FFFF00';
var g_seg_red    = '#FF0000';
var g_seg_colors; // Cache mappings of segment indices to colors

// Cache map of dBFS to segment number
var g_val_to_seg_cache = new Array(1000);

// 'vertical' or 'horizontal'
var g_mtr_layout = 'horizontal';

// Regular expressions for parsing lwrp
// Grammar and official regex implementations are defined at: 
// http://axia.lv/wiki/index.php/Livewire_Routing_Protocol#Syntax
// g_lwrp_token: Break lwrp line into seperate tokens 
var g_lwrp_token = /([A-Z]+:("(\\.|[^"\\])*"|(\\.|[^\s"\\])*)|\S+)/g;
// g_lwrp_strip1 and g_lwrp_strip2: Remove unnecessary quotes and escape characters from a value before displaying
var g_lwrp_strip1 = /(^"|"$)/g; 
var g_lwrp_strip2 = /\\(?=[\\"\s])/g; 

// Helper function for error logging
String.prototype.elemNotFound = function () {
	return "Element '" + this + "' not found!";
}

function element_set_display2 (elem, display)
{
	var val = display ? "block" : "none";
	elem && (elem.style.display != val) && (elem.style.display = val);
}

function element_set_display (id, display)
{
	element_set_display2(document.getElementById(id), display);
}

// get_request_obj(): Get appropriate object to use for AJAX requests, depending on browser
function get_request_obj()
{
	var req;
	try {  
		req = new window.XDomainRequest();
	} 
	catch(e) {}
	if(!req) {
		try { 
			req = new XMLHttpRequest(); 
		} 
		catch(e) {} 
	}
	return req;
}

// load_java_meters(): Load legacy java meters applet if meters.js fails
function load_java_meters()
{
	var msg  = "Using legacy meters. For improved meters, update your browser to the latest version of Chrome, Firefox, Safari, or Internet Explorer.";
	var html = "<br>" + msg + "<br><br><br><applet codebase=\"/classes\" code=\"LvMtr.class\" width=\"100%\" height=\"400\"></applet>";

	//console.log(msg);
	g_mtr_element.innerHTML = html;
	return;
}

// meters_init(): Start meters
//   elem: ID of element in which to draw meters (sets innerHTML of this element)
function meters_init(elem, layout, segments)
{
	g_mtr_element = document.getElementById(elem);
	if (!g_mtr_element) {
		var msg = id.elemNotFound();
		//console.log(msg);
		document.write(msg);
		return;
	}

	if (layout) {
		g_mtr_layout = layout;
	}

	if (segments) {
		g_nsegments = segments;
	}

	g_mtr_element.innerHTML = "<span>" 
	 //+ "<br>Meters Loading...<br>"
	 + "<div id='meters_load_error'></div>"
	 + "</span>";

	// If meters don't load after 3 seconds, show error message
	setTimeout(meters_load_error_msg, 3000);
	setTimeout(meters_init2, 1);
}

// meters_load_error_msg(): Will print error message if meters have not yet loaded
function meters_load_error_msg()
{
	var elem = document.getElementById('meters_load_error');
	if (!elem) return;
	elem.innerHTML = 
	   "<br>If meters fail to start:<br>"
	 + " - Update your browser to the latest version of Google Chrome, Mozilla Firefox, Apple Safari or Microsoft Internet Explorer.<br>"
}

// meters_init2(): Start AJAX connection, called from setTimeout to make browser think page load is finished
function meters_init2()
{
	// Initialize cache elements that don't depend on meters_draw()

	// Compile the regex for improved speed
	g_lwrp_token.compile(g_lwrp_token); 
	g_lwrp_strip1.compile(g_lwrp_strip1); 
	g_lwrp_strip2.compile(g_lwrp_strip2); 

	// Fake response, because we're not really using livewire meters anyway
	//lwr_connect("VER");
	lwr_process('VER NSRC:1 NDST:1');
	lwr_process('SRC 1 NAME:"Xmt" NCHN:2');
	lwr_process('DST 1 NAME:"Rcv" NCHN:2');
}

// lwr_process(): Process one line of lwrp data
// XXX: Does not process meter values
function lwr_process(line)
{
	//console.log("lwr_process() line = " + line);
	var tokens = line.match(g_lwrp_token);
	if (!tokens) {
		return;
	}
	var cmd = tokens[0];
	switch (cmd) {
		case 'SRC':
		case 'DST':
			var cmdl = cmd.toLowerCase();
			var chan = parseInt(tokens[1]);
			if (!chan) break;

			var new_nchn = 2; // If NCHN not set, assume stereo
			for (var i in tokens) {
				var nameval = tokens[i];
				if (typeof nameval!="string") {
					continue;
				}
				var seperator = nameval.indexOf(':');
				if (seperator < 0) {
					continue;
				}
				var prop = nameval.slice(0, seperator);
				var val  = nameval.slice(seperator+1);
				switch (prop) {
					case 'PSNM': 
					case 'NAME': 
						var flags = (g_surround[cmdl] ? "nodraw" : "");
						meters_set_label(cmdl, chan, val, flags);
						break;
					case 'NCHN':
						var tmp = parseInt(val);
						if (!isNaN(tmp)) new_nchn = tmp;
						break;
				}
				if (new_nchn != g_nchn[cmdl][chan]) {
					meters_set_nchn(cmdl, chan, new_nchn);
					// Toggle visibility of meters
					var visible = (new_nchn!=0);
					var src_and_chan = cmdl + "_" + chan;

					g_nchn[cmdl][chan] = new_nchn;
					element_set_display(src_and_chan + '_mtr', visible);
					element_set_display(src_and_chan + '_label', visible);
				}
			}
			break;
		case 'VER':
			g_VER = line;
			var nsrc = g_nsrc;
			var ndst = g_ndst;
			for (var i in tokens) {
				var nameval = tokens[i];
				if (typeof nameval!="string") {
					continue;
				}
				var seperator = nameval.indexOf(':');
				if (seperator < 0) {
					continue;
				}
				var prop = nameval.slice(0, seperator);
				var val  = nameval.slice(seperator+1);
				switch (prop) {
					case 'NSRC':
						var tmp = parseInt(val);
						if (!isNaN(tmp)) nsrc = tmp;
						break;
					case 'NDST':
						var tmp = parseInt(val);
						if (!isNaN(tmp)) ndst = tmp;
						break;
				}
			}

			if (nsrc!=g_nsrc || ndst!=g_ndst) {
				g_mtr_request && g_mtr_request.abort();

				switch (g_mtr_layout) {
					case 'vertical':
						meters_draw_vertical(nsrc, ndst);
						break;
					case 'horizontal':
					default:
						meters_draw_horizontal(nsrc, ndst);
						break;
				}
				meters_connect();
			}
			break;
		case 'MTR':
			//meters_process(line);
			break;
	}
}

// meters_set_nchn(): Set number of audio channels (off, mono, or stereo) for lw channel
function meters_set_nchn(src_or_dst, chan, new_nchn)
{
	// Swap stereo and mono meters
	var src_and_chan = src_or_dst + "_" + chan;
	var e = document.getElementById(src_and_chan + '_mtr');
	var e2 = g_mtr_fragments[src_or_dst][chan][new_nchn];
	if (!e || !e2) return;
	if (e.hasChildNodes()) {
		e.replaceChild(e2, e.firstChild);
	}
	else {
		e.appendChild(e2);
	}
}

// meters_set_label(): Set new label for channel
//   label: new channel label, null refreshes prevVal
//   flags: can contain any of the following string values
//     - "force": force change, but do not cache value
//     - "nodraw": cache, but do not draw change
//     - "noproc": do not strip, split, or join value
function meters_set_label(src_or_dst, chan, label, flags)
{
	var id = src_or_dst + "_" + chan + "_label";
	var e = document.getElementById(id);
	if (e) {
		// Hide the previous pre-processed value in e to optimize comparisons
		if (!label) {
			label = e.prevVal;
		}
		else if (!/force/.test(flags)) {
			if (label == e.prevVal) {
				return;
			}
			e.prevVal = label; 
		}
		if (/nodraw/.test(flags)) return;
		if (!(/noproc/.test(flags))) {
			label = label.replace(g_lwrp_strip1,'').replace(g_lwrp_strip2,''); // Do strip in two stages
		}
		e.innerHTML = label;
	}
	else {
		//console.log(id.elemNotFound());
	}
}

// meters_draw_horizontal(): Draw meter table, should only need to be called on startup
//   nsrc: Number of source meters
//   ndst: Number of destination meters
function meters_draw_horizontal(nsrc, ndst) 
{
	var html = "<table class=\"meters\" style='padding: 4px 0px 4px 0px;'><tbody>";

	// Draw columns for src or dst table
	var draw_meter_columns = function (src_or_dst, channels) {
		var html = "";
		var nchan = channels.size;
		var cache = (g_mtr_fragments[src_or_dst] = new Array(nchan));

		var L_elems;
		var R_elems;
		var M_elems;
		if (src_or_dst=='src') {
			L_elems = (g_src_L_elems = new Array(nchan));
			R_elems = (g_src_R_elems = new Array(nchan));
			M_elems = (g_src_M_elems = new Array(nchan));
		}
		else if (src_or_dst=='dst') {
			L_elems = (g_dst_L_elems = new Array(nchan));
			R_elems = (g_dst_R_elems = new Array(nchan));
			M_elems = (g_dst_M_elems = new Array(nchan));
		}

		for (var i in channels) {
			var chan = channels[i];
			var cache2 = (cache[chan] = new Array());
			var L_elems2 = (L_elems[chan] = new Array(g_nsegments));
			var R_elems2 = (R_elems[chan] = new Array(g_nsegments));
			var M_elems2 = (M_elems[chan] = new Array(g_nsegments));
			var src_and_chan = src_or_dst + "_" + chan;

			// Create container for meters
			html += "<td class='nopad'><div class='nodisplay' style='padding: 2px 0px 2px 3px;' id='" + src_and_chan + "_mtr'></div></td>";

			// Create tables mono and stereo meters, but do not show yet
			// They will be kept as document fragments and added later when NCHN is read
			var table = document.createElement("table");
			table.className = "hmeter";

			var tr1 = document.createElement("tr");
			for (var j=0; j<g_nsegments; j++) {
				var td = document.createElement("td");
				td.className = "hmeter_segment";
				td.id = src_and_chan + "_L_" + j;
				L_elems2[j] = td;
				tr1.appendChild(td);
			}
			table.appendChild(tr1);
			var tr2 = document.createElement("tr");
			for (var j=0; j<g_nsegments; j++) {
				var td = document.createElement("td");
				td.className = "hmeter_segment";
				td.id = src_and_chan + "_L_" + j;
				R_elems2[j] = td;
				tr2.appendChild(td);
			}
			table.appendChild(tr2);
			cache2[2] = table;
		}
		return html;
	};

	var src_chans = new Array();
	var dst_chans = new Array();

	for (var i=1; i<=nsrc; i++) {
		src_chans[i] = i;
	}
	for (var i=1; i<=ndst; i++) {
		dst_chans[i] = i;
	}

	// Draw meters
	html += "<tr>";
	html += "<td class='bold'>Xmt</td>";
	html += draw_meter_columns('src', src_chans);
	html += "</tr>";

	html += "<tr>";
	html += "<td class='bold'>Rcv</td>";
	html += draw_meter_columns('dst', dst_chans);
	html += "</tr>";

	// Area for displaying error messages
	html += "<tr><td class='center' id='meter_msg' colspan='100%'></td></tr>";
	html += "</tbody></table>";

	g_mtr_element.innerHTML = html;
	g_nsrc = nsrc;
	g_ndst = ndst;

	meters_create_cache();
};

// meters_draw_vertical(): Draw meter table, should only need to be called on startup
//   nsrc: Number of source meters
//   ndst: Number of destination meters
function meters_draw_vertical(nsrc, ndst) 
{
	var html = "<table class=\"meters\" style='padding: 0px 16px 0px 16px;'><tbody>";

	// Draw columns for src or dst table
	var draw_meter_columns = function (src_or_dst, channels) {
		var html = "";
		var nchan = channels.size;
		var cache = (g_mtr_fragments[src_or_dst] = new Array(nchan));

		var L_elems;
		var R_elems;
		var M_elems;
		if (src_or_dst=='src') {
			L_elems = (g_src_L_elems = new Array(nchan));
			R_elems = (g_src_R_elems = new Array(nchan));
			M_elems = (g_src_M_elems = new Array(nchan));
		}
		else if (src_or_dst=='dst') {
			L_elems = (g_dst_L_elems = new Array(nchan));
			R_elems = (g_dst_R_elems = new Array(nchan));
			M_elems = (g_dst_M_elems = new Array(nchan));
		}

		for (var i in channels) {
			var chan = channels[i];
			var cache2 = (cache[chan] = new Array());
			var L_elems2 = (L_elems[chan] = new Array(g_nsegments));
			var R_elems2 = (R_elems[chan] = new Array(g_nsegments));
			var M_elems2 = (M_elems[chan] = new Array(g_nsegments));
			var src_and_chan = src_or_dst + "_" + chan;

			// Create container for meters
			html += "<td class='nopad'><div class='nodisplay' style='padding: 0px 0px 4px 0px;' id='" + src_and_chan + "_mtr'></div></td>";

			// Create tables mono and stereo meters, but do not show yet
			// They will be kept as document fragments and added later when NCHN is read
			var table1 = document.createElement("table");
			table1.className = "meter meter_col_width";
			var table2 = table1.cloneNode(true);

			// Internet Exploder REQUIRES these
			var tbody1 = document.createElement("tbody");
			var tbody2 = document.createElement("tbody");
			
			for (var j=g_nsegments; j-->0;) {
				var td_L = document.createElement("td");
				td_L.className = "meter_segment halfwide";
				var td_R = td_L.cloneNode(true);

				var td_M = document.createElement("td");
				td_M.className = "meter_segment fullwide";

				td_L.id = src_and_chan + "_L_" + j;
				td_R.id = src_and_chan + "_R_" + j;
				td_M.id = src_and_chan + "_M_" + j;

				var tr_LR = document.createElement("tr");
				var tr_M = tr_LR.cloneNode(true);

				L_elems2[j] = td_L;
				R_elems2[j] = td_R;
				M_elems2[j] = td_M;

				tr_M.appendChild(td_M);
				tr_LR.appendChild(td_L);
				tr_LR.appendChild(td_R);

				tbody1.appendChild(tr_M);
				tbody2.appendChild(tr_LR);
			}
			table1.appendChild(tbody1);
			table2.appendChild(tbody2);
			cache2[1] = table1;
			cache2[2] = table2;
		}
		return html;
	};
	var draw_labels = function (src_or_dst, channels) {
		var html = "";
		for (var i in channels) {
			var chan = channels[i];
			var id = src_or_dst + "_" + chan + "_label";
			html += "<td class='nopad topalign'><div class='nodisplay small bold center' id='" + id + "'>";
			html += chan;
			html += "</div></td>";
		}
		return html;
	};

	var src_chans = new Array();
	var dst_chans = new Array();

	// Will be set to true if meters require atypical arrangement
	var src_special = false;
	var dst_special = false;

	for (var i=1; i<=nsrc; i++) {
		src_chans[i] = i;
	}
	for (var i=1; i<=ndst; i++) {
		dst_chans[i] = i;
	}

	// Draw meters
	html += "<tr>";
	html += draw_meter_columns('src', src_chans);
	html += "<td class='meter_gap_width'></td>"; // Gap column
	html += draw_meter_columns('dst', dst_chans);
	html += "</tr>";

	// Draw meter labels
	html += "<tr>";
	html += draw_labels('src', src_chans);
	html += "<td class='meter_gap_width'></td>"; // Gap column
	html += draw_labels('dst', dst_chans);
	html += "</tr>";

	// Area for displaying error messages
	html += "<tr><td class='center' id='meter_msg' colspan='100%'></td></tr>";
	html += "</tbody></table>";

	g_mtr_element.innerHTML = html;
	g_nsrc = nsrc;
	g_ndst = ndst;

	meters_create_cache();
};

// meters_create_cache(): Precompute and cache variables for use in meter update functions
//   Variables here are those that depend on meters_draw()
function meters_create_cache()
{
	// Create arrays for caching segment counts, and set to zero
	var size = g_nsrc+1;
	g_src_L_seg         = new Array(size);
	g_src_R_seg         = new Array(size);
	g_src_M_seg         = new Array(size);
	g_nchn['src']       = new Array(size);
	for (var chan=1; chan<size; chan++) {
		g_src_L_seg[chan]         = 0;
		g_src_R_seg[chan]         = 0;
		g_src_M_seg[chan]         = 0;
		g_nchn['src'][chan]       = 0;
	}

	size = g_ndst+1;
	g_dst_L_seg         = new Array(size);
	g_dst_R_seg         = new Array(size);
	g_dst_M_seg         = new Array(size);
	g_nchn['dst']       = new Array(size);
	for (var chan=1; chan<size; chan++) {
		g_dst_L_seg[chan]         = 0;
		g_dst_R_seg[chan]         = 0;
		g_dst_M_seg[chan]         = 0;
		g_nchn['dst'][chan]       = 0;
	}

	// Create cache of meter colors
	g_seg_colors = new Array(g_nsegments);

	// Create cache of all possible values passed to meters_val_to_seg()
	// and create cache table of meter segment elements
	// and set dBFS labels on meter segments
	var last_seg_id = null;
	for (var i=0; i<=1000; i++) {
		var dbfs10 = -i;
		var seg = meters_val_to_seg(dbfs10);
		g_val_to_seg_cache[i] = seg;
		//console.log("meters_val_to_seg(" + dbfs10 + ")=" + seg);
		var seg_id = seg - 1;
		if (seg_id!=last_seg_id && seg_id >= 0) {
			// Compute segment colors
			var color;
			if      (dbfs10 > -80)  color = g_seg_red;
			else if (dbfs10 > -200) color = g_seg_yellow;
			else                    color = g_seg_green;
			g_seg_colors[seg_id] = color;

			var labelNum = dbfs10/10;
			if (dbfs10 < -250) {
				labelNum = Math.floor(labelNum);
			}
			var label = labelNum + ' dBFS';
			for (var chan=1; chan<=g_nsrc; chan++) {
				g_src_L_elems[chan][seg_id].title = label;
				g_src_R_elems[chan][seg_id].title = label;
			}
			for (var chan=1; chan<=g_ndst; chan++) {
				g_dst_L_elems[chan][seg_id].title = label;
				g_dst_R_elems[chan][seg_id].title = label;
			}
			last_seg_id = seg_id;
		}
	}
}

// meters_timeout(): Called when meters connection times out
function meters_timeout () 
{
	var msg = 'Connection Lost!';
	meters_set_msg(msg, 'red');
	//console.log(msg); 
	g_mtr_conn_err = true;
	g_mtr_request.abort();

	// Limit max reconnect delay to 30 seconds
	g_mtr_reconnect = Math.min(g_mtr_reconnect*2, 30000);
	
	clearTimeout(g_mtr_timeout);
	g_mtr_timeout = setTimeout(meters_connect, g_mtr_reconnect);
}

// meters_connect(): Initiate ajax connection
function meters_connect()
{
	var url = "/lwr/mtr?duration=-1&interval=" + g_mtr_interval + "&sid=" + Date.now();
	//console.log("meters_connect(): url = " + url);

	g_mtr_request && g_mtr_request.abort();
	g_mtr_request = get_request_obj();
	g_mtr_request.open("GET", url, true);
	g_mtr_request.onprogress = meters_recv;
	g_mtr_request.send(null);

	g_mtr_resp_idx = 0; 
	clearTimeout(g_mtr_timeout);
	g_mtr_timeout = setTimeout(meters_timeout, g_mtr_timeout_ms);
}

// meters_recv(): Revcieve incoming data from xmlHttpRequest 
function meters_recv()
{
	//console.log("meters_recv()");

	if (g_mtr_request.readyState < 3) {
		return;
	}

	// Reset timeout timer
	g_mtr_reconnect = 1000;
	clearTimeout(g_mtr_timeout);
	g_mtr_timeout = setTimeout(meters_timeout, g_mtr_timeout_ms);

	var response = g_mtr_request.responseText.substring(g_mtr_resp_idx);
	if (!response) {
		return;
	}
	g_mtr_resp_idx += response.lastIndexOf('\n') + 1;

	if (g_mtr_conn_err) {
		meters_set_msg('&nbsp;', 'green'); // Erase error message
		g_mtr_conn_err = false;
	}

	var lines = response.split('\n');
	for (var i=0; i < lines.length; i++) {
		meters_process(lines[i]);
	}

	if (g_mtr_request.responseText.length > g_resp_idx_limit) {
		// If message data gets too big, reconnect
		meters_connect();
	}
}

// meters_process(): Process one line of meter data
function meters_process(line)
{
	// Sample line: MTR ICH 1 PEEK:-1000:-1000 RMS:-1000:-1000
	var tokens = line.split(' ');

	if (tokens.length < 4) return;
	if (tokens[0] != 'MTR') return;

	var peek = tokens[3].split(':');
	if (peek.length < 3) return;

	var chan   = tokens[2];
	var peek_L = parseInt(peek[1]);
	var peek_R = parseInt(peek[2]);

	switch (tokens[1]) {
		case 'ICH' : 
			switch (g_nchn['dst'][chan]) {
				case 1:
					g_dst_M_seg[chan] = meters_fill_segments(g_dst_M_elems[chan], peek_L, g_dst_M_seg[chan]);
					break;
				case 2:
					g_dst_L_seg[chan] = meters_fill_segments(g_dst_L_elems[chan], peek_L, g_dst_L_seg[chan]);
					g_dst_R_seg[chan] = meters_fill_segments(g_dst_R_elems[chan], peek_R, g_dst_R_seg[chan]);
					break;
			}
			break;
		case 'OCH' : 
			switch (g_nchn['src'][chan]) {
				case 1:
					g_src_M_seg[chan] = meters_fill_segments(g_src_M_elems[chan], peek_L, g_src_M_seg[chan]);
					break;
				case 2:
					g_src_L_seg[chan] = meters_fill_segments(g_src_L_elems[chan], peek_L, g_src_L_seg[chan]);
					g_src_R_seg[chan] = meters_fill_segments(g_src_R_elems[chan], peek_R, g_src_R_seg[chan]);
					break;
			}
			break;
		default: 
			return;
	}
}

// meters_fill_segments(): Color segments on a meter
//   id:       element id NOT INCLUDING last index of meter segment
//   dbfs10:   value of meter in 1/10 dBFS
//   last_seg: number of segments filled on this meter last time
// returns: number of segments drawn
// XXX: This is called >1000 times/second, so must be fast!!!
//      Pre-compute and cache as much as possible!
function meters_fill_segments(elems, dbfs10, last_seg) 
{
	//var segments = meters_val_to_seg(dbfs10);
	var segments = g_val_to_seg_cache[-dbfs10] || 0;

	if (segments > last_seg) {
		var max_change = last_seg + 3;
		if (segments > max_change) {
			segments = max_change;
		}
		for (var i=last_seg; i<segments; i++) {
			elems[i].style.backgroundColor = g_seg_colors[i];
		}
	}
	else if (segments < last_seg) {
		var max_change = last_seg - 2;
		if (segments < max_change) {
			segments = max_change;
		}
		for (var i=segments; i<last_seg; i++) {
			elems[i].style.backgroundColor = g_seg_black;
		}
	}

	// return this so meters_process() can cache it and send it again next time
	return segments;
}

// meters_val_to_seg(): Convert 1/10 dBFS value to # of meter segments
//   dbfs10: meter value in units 1/10 dBFS
// returns: number of segments to draw 
function meters_val_to_seg(dbfs10)
{
	if(dbfs10 <= -1000)
		return 0;

	var dbfs = dbfs10 / 10;

	// This part is taken from an old java applet.
	// Can give results from 0 (0 dbFS) to -32 (-100 dBFS)
	// Rescale to 0..1
	var tmp;
	if(dbfs >= -24)
		tmp = dbfs;
	else if(dbfs >= -26)
		tmp = -24 + (dbfs + 24)/2;
	else if(dbfs >= -30)
		tmp = -25 + (dbfs + 26)/4;
	else if(dbfs >= -60)
		tmp = -26 + (dbfs + 30)/6;
	else 
		tmp = -30 + (dbfs + 60)/20;

	tmp /= 32;
	tmp += 1;

	// Scale to how many segments we actually have
	var seg = tmp * g_nsegments;

	return Math.floor(seg);
}

// meters_set_msg(): Display error messages
// XXX: Don't call too often, for some reason, this is VERY slow on firefox
function meters_set_msg(msg, color) 
{
	var e = document.getElementById('meter_msg');
	if (e) {
		e.innerHTML   = msg;
		e.style.color = color;
	}
	else {
		//console.log('meter_msg'.elemNotFound());
	}
}
